@@ -1,8 +1,106 @@
 """ Common functions for package biobb_cmip.cmip """
 import re
+import json
 from pathlib import Path
-from typing import List, Dict, Tuple, Mapping, Union, Set, Sequence
-import pytraj as pt
+from typing import List, Dict, Mapping, Union, Tuple, Sequence
+import MDAnalysis as mda
+from MDAnalysis.topology.guessers import guess_atom_element
+import uuid
+import logging
+import biobb_common.tools.file_utils as fu
+
+
+def get_grid(cmip_log_path: Union[str, Path]) -> Tuple[Tuple[float, float, float], Tuple[float, float, float], Mapping[str, Tuple[float, float, float]]]:
+    with open(cmip_log_path) as log_file:
+        first_line = log_file.readline().strip()
+    if first_line.startswith("titleParam"):
+        return _get_grid_from_key_value(cmip_log_path)
+    elif first_line.startswith("{"):
+        return _get_grid_from_box_file(cmip_log_path)
+    return _get_grid_from_text(cmip_log_path)
+
+
+def _get_grid_from_box_file(cmip_box_path: Union[str, Path]) -> Tuple[Tuple[float, float, float], Tuple[float, float, float], Mapping[str, Tuple[float, float, float]]]:
+    with open(cmip_box_path) as json_file:
+        grid_dict = json.load(json_file)
+    origin = grid_dict['origin']['x'], grid_dict['origin']['y'], grid_dict['origin']['z']
+    size = grid_dict['size']['x'], grid_dict['size']['y'], grid_dict['size']['z']
+    return origin, size, grid_dict['params']
+
+
+def _get_grid_from_text(cmip_log_path: Union[str, Path]) -> Tuple[Tuple[float, float, float], Tuple[float, float, float], Mapping[str, Tuple[float, float, float]]]:
+    origin = None
+    size = None
+    grid_params = {"CEN": None, "DIM": None, "INT": None}
+
+    with open(cmip_log_path) as log_file:
+        inside_automatic_grid = False
+        for line in log_file:
+            if line.strip() in ["AUTOMATIC GRID", "MANUAL GRID"]:
+                inside_automatic_grid = True
+            if inside_automatic_grid:
+                origin_match = re.match(r"Grid origin:\s+([-+]?(?:\d*\.\d+|\d+))\s+([-+]?(?:\d*\.\d+|\d+))\s+([-+]?(?:\d*\.\d+|\d+))", line.strip())
+                if origin_match:
+                    origin = float(origin_match.group(1)), float(origin_match.group(2)), float(origin_match.group(3))
+                size_match = re.match(r"Grid Size:\s+([-+]?(?:\d*\.\d+|\d+))\s+([-+]?(?:\d*\.\d+|\d+))\s+([-+]?(?:\d*\.\d+|\d+))", line.strip())
+                if size_match:
+                    size = float(size_match.group(1)), float(size_match.group(2)), float(size_match.group(3))
+                int_match = re.match(r"Grid units:\s+([-+]?(?:\d*\.\d+|\d+))\s+([-+]?(?:\d*\.\d+|\d+))\s+([-+]?(?:\d*\.\d+|\d+))", line.strip())
+                if int_match:
+                    grid_params['INT'] = float(int_match.group(1)), float(int_match.group(2)), float(int_match.group(3))
+                cen_match = re.match(r"Grid center:\s+([-+]?(?:\d*\.\d+|\d+))\s+([-+]?(?:\d*\.\d+|\d+))\s+([-+]?(?:\d*\.\d+|\d+))",line.strip())
+                if cen_match:
+                    grid_params['CEN'] = float(cen_match.group(1)), float(cen_match.group(2)), float(cen_match.group(3))
+                dim_match = re.match(r"Grid density:\s+([-+]?(?:\d*\.\d+|\d+))\s+([-+]?(?:\d*\.\d+|\d+))\s+([-+]?(?:\d*\.\d+|\d+))", line.strip())
+                if dim_match:
+                    grid_params['DIM'] = int(dim_match.group(1)), int(dim_match.group(2)), int(dim_match.group(3))
+                if origin and size and grid_params['INT'] and grid_params['CEN'] and grid_params['DIM']:
+                    break
+    return origin, size, grid_params
+
+
+def _get_grid_from_key_value(cmip_log_path: Union[str, Path]) -> Tuple[Tuple[float, float, float], Tuple[float, float, float], Mapping[str, Tuple[float, float, float]]]:
+    origin = None
+    size = None
+    grid_params = {"CEN": None, "DIM": None, "INT": None}
+
+    with open(cmip_log_path) as log_file:
+        inside_automatic_grid = False
+        for line in log_file:
+            if line.strip() in ["AUTOMATIC GRID", "MANUAL GRID"]:
+                inside_automatic_grid = True
+            if inside_automatic_grid:
+                origin_match = re.match(r"origin=\s+([-+]?(?:\d*\.\d+|\d+))\s*,\s*([-+]?(?:\d*\.\d+|\d+))\s*,\s*([-+]?(?:\d*\.\d+|\d+))", line.strip())
+                if origin_match:
+                    origin = float(origin_match.group(1)), float(origin_match.group(2)), float(origin_match.group(3))
+                size_match = re.match(r"size=\s+([-+]?(?:\d*\.\d+|\d+))\s*,\s*([-+]?(?:\d*\.\d+|\d+))\s*,\s*([-+]?(?:\d*\.\d+|\d+))", line.strip())
+                if size_match:
+                    size = float(size_match.group(1)), float(size_match.group(2)), float(size_match.group(3))
+                int_match = re.match(r"spacing=\s+([-+]?(?:\d*\.\d+|\d+))\s*,\s*([-+]?(?:\d*\.\d+|\d+))\s*,\s*([-+]?(?:\d*\.\d+|\d+))", line.strip())
+                if int_match:
+                    grid_params['INT'] = float(int_match.group(1)), float(int_match.group(2)), float(int_match.group(3))
+                cen_match = re.match(r"center=\s+([-+]?(?:\d*\.\d+|\d+))\s*,\s*([-+]?(?:\d*\.\d+|\d+))\s*,\s*([-+]?(?:\d*\.\d+|\d+))",line.strip())
+                if cen_match:
+                    grid_params['CEN'] = float(cen_match.group(1)), float(cen_match.group(2)), float(cen_match.group(3))
+                dim_match = re.match(r"dim=\s+([-+]?(?:\d*\.\d+|\d+))\s*,\s*([-+]?(?:\d*\.\d+|\d+))\s*,\s*([-+]?(?:\d*\.\d+|\d+))", line.strip())
+                if dim_match:
+                    grid_params['DIM'] = int(dim_match.group(1)), int(dim_match.group(2)), int(dim_match.group(3))
+                if origin and size and grid_params['INT'] and grid_params['CEN'] and grid_params['DIM']:
+                    break
+    return origin, size, grid_params
+
+
+
+def create_unique_file_path(parent_dir: Union[str, Path] = None, extension: str = None) -> str:
+    if not parent_dir:
+        parent_dir = Path.cwd
+    if not extension:
+        extension = ''
+    while True:
+        name = str(uuid.uuid4())+extension
+        file_path = Path.joinpath(Path(parent_dir).resolve(), name)
+        if not file_path.exists():
+            return str(file_path)
 
 
 def write_cmip_pdb(input_pdb_path, output_pdb_path, charges_list, elements_list):
@@ -15,6 +113,7 @@ def write_cmip_pdb(input_pdb_path, output_pdb_path, charges_list, elements_list)
             outPDB.write("{}{:8.4f}  {}\n".format(line[:54], charges_list[index], elements_list[index]))
             index += 1
 
+
 def get_topology_cmip_elements_canonical(input_topology_filename: str) -> List:
     """
     This function also accepts pdb files
@@ -24,53 +123,32 @@ def get_topology_cmip_elements_canonical(input_topology_filename: str) -> List:
     Returns:
 
     """
-    topology = pt.load_topology(filename=input_topology_filename)
-
-    # Set all supported elements
-    # This is required to transform element names (returned by pytraj) to element letters
-    standard_elements = {
-        'hydrogen': 'H',
-        'carbon': 'C',
-        'oxygen': 'O',
-        'nitrogen': 'N',
-        'sulfur': 'S',
-        'sodium': 'Na',
-        'chlorine': 'Cl',
-        'zinc': 'Zn',
-        'fluorine': 'F',
-        'magnesium': 'Mg',
-        'phosphorus': 'P',
-    }
-    # Iterate over each atom to save their CMIP element
-    elements = []
-    atoms = list(topology.atoms)
-    for a, atom in enumerate(atoms):
-        residue = atom.resname
-        # Skip this atom if we already found it
-        name = atom.name
-        element = standard_elements[atom.element]
-        # Adapt hydrogens element to CMIP requirements
-        if element == 'H':
-            # There should we always only 1 bond
-            # If you have the error below you may need to updated the pytraj version or reintsall pytraj
-            # ValueError: Buffer dtype mismatch, expected 'int' but got 'long'
-            bonded_heavy_atom_index = atom.bonded_indices()[0]
-            bonded_heavy_atom = atoms[bonded_heavy_atom_index]
-            bonded_heavy_atom_element = standard_elements[bonded_heavy_atom.element]
-            # Hydrogens bonded to carbons remain as 'H'
-            if bonded_heavy_atom_element == 'C':
-                pass
-            # Hydrogens bonded to oxygen are renamed as 'HO'
-            elif bonded_heavy_atom_element == 'O':
-                element = 'HO'
-            # Hydrogens bonded to nitrogen or sulfur are renamed as 'HN'
-            elif bonded_heavy_atom_element == 'N' or bonded_heavy_atom_element == 'S':
-                element = 'HN'
-            else:
-                raise SystemExit(
-                    'ERROR: Hydrogen bonded to not supported heavy atom: ' + bonded_heavy_atom_element)
-        elements.append(element)
-    return elements
+    # Remove forcefield itp references from top file.
+    if input_topology_filename.lower().endswith('.top'):
+        with open(input_topology_filename) as tf:
+            top_lines = tf.readlines()
+        top_file = create_unique_file_path(parent_dir=Path(input_topology_filename).parent.resolve(), extension='.top')
+        with open(top_file, 'w') as nt:
+            for line in top_lines:
+                if re.search(r"\.ff.*\.itp", line):
+                    continue
+                nt.write(line)
+        u = mda.Universe(top_file, topology_format="ITP")
+    else:
+        u = mda.Universe(input_topology_filename)
+    # mda_charges = [round(val, 4) for val in u.atoms.charges]
+    # mda_atom_types = list(guess_types(u.atoms.names))
+    mda_atom_types = []
+    for atom in u.atoms:
+        atom_element = guess_atom_element(atom.name)
+        if atom_element == 'H':
+            bonded_atom_element = guess_atom_element(atom.bonded_atoms[0].name)
+            if bonded_atom_element == 'O':
+                atom_element = 'HO'
+            elif bonded_atom_element in ['N', 'S']:
+                atom_element = 'HN'
+        mda_atom_types.append(atom_element)
+    return mda_atom_types
 
 
 def get_topology_charges(input_topology_filename: str) -> List:
@@ -78,10 +156,20 @@ def get_topology_charges(input_topology_filename: str) -> List:
     Extract those charges and save them in a list to be returned
     Supported formats (tested): prmtop, top, psf
     """
-    topology = pt.load_topology(filename=input_topology_filename)
-    # WARNING: We must convert this numpy ndarray to a normal list otherwise the search by index is extremely inefficient
-    topology_charges = list(topology.charge)
-    return topology_charges
+    # Remove forcefield itp references from top file.
+    if input_topology_filename.lower().endswith('.top'):
+        with open(input_topology_filename) as tf:
+            top_lines = tf.readlines()
+        top_file = create_unique_file_path(parent_dir=Path(input_topology_filename).parent.resolve(), extension='.top')
+        with open(top_file, 'w') as nt:
+            for line in top_lines:
+                if re.search(r"\.ff.*\.itp", line):
+                    continue
+                nt.write(line)
+        u = mda.Universe(top_file, topology_format="ITP")
+    else:
+        u = mda.Universe(input_topology_filename)
+    return [round(val, 4) for val in u.atoms.charges]
 
 
 class Residue:
@@ -206,6 +294,7 @@ def probe_params_grid(probe_id: int = 0, readgrid: int = 2, pbfocus: int = 1, pe
 
     return grid_dict
 
+
 def params_grid(grid_type: str = 'cmip', readgrid: int = 0, perfill: float = 0.8,
                 grid_int: Sequence[float] = (0.5, 0.5, 0.5),
                 grid_dim: Sequence[float] = (64, 64, 64),
@@ -247,20 +336,50 @@ def params_preset(execution_type: str) -> Dict[str, str]:
             'titration': 1, 'inifoc': 2, 'cutfoc': -0.5, 'focus': 1, 'ninter': 10, 'clhost': 1, 'titcut': 20.,
             'titwat': 10, 'titip': 10, 'titim': 10
         }
-    elif execution_type == 'mip':
-        grid_dict = params_grid(grid_type='mip', readgrid= 0,
-                                grid_cen=(47.266, 83.4265, 54.174),
-                                grid_dim=(46, 30, 37),
-                                grid_int=(0.5, 0.5, 0.5))
+    elif execution_type.strip() == 'mip_pos':
+        grid_dict = params_grid(grid_type='mip', readgrid=2)
         params_dict = {
-            'title': 'MIP O-  Mehler Solmajer dielectric',
+            'title': 'MIP positive probe',
             'tipcalc': 0,
             'calcgrid': 1,
             'irest': 0,
             'orest': 0,
             'coorfmt': 2,
             'dields':  2,
-            'cubeoutput': 1, 'carmip': -1, 'fvdw': 0.8
+            'cubeoutput': 1,
+            'fvdw': 0.8,
+            'carmip': 1,
+            'tipatmip' : "'OW'"
+        }
+    elif execution_type.strip() == 'mip_neu':
+        grid_dict = params_grid(grid_type='mip', readgrid=2)
+        params_dict = {
+            'title': 'MIP neutral probe',
+            'tipcalc': 0,
+            'calcgrid': 1,
+            'irest': 0,
+            'orest': 0,
+            'coorfmt': 2,
+            'dields':  2,
+            'cubeoutput': 1,
+            'fvdw': 0.8,
+            'carmip': 0,
+            'tipatmip' : "'OW'"
+        }
+    elif execution_type.strip() == 'mip_neg':
+        grid_dict = params_grid(grid_type='mip', readgrid=2)
+        params_dict = {
+            'title': 'MIP negative probe',
+            'tipcalc': 0,
+            'calcgrid': 1,
+            'irest': 0,
+            'orest': 0,
+            'coorfmt': 2,
+            'dields':  2,
+            'cubeoutput': 1,
+            'fvdw': 0.8,
+            'carmip': -1,
+            'tipatmip' : "'OW'"
         }
 #TODO 'carmip': 1,
     # wat: tipcalc: 1 + titration: 'inifoc': 2, 'cutfoc': -0.5, 'focus': 1, 'ninter': 10,
@@ -279,10 +398,7 @@ def params_preset(execution_type: str) -> Dict[str, str]:
         }
 
     elif execution_type == 'energy':
-        grid_dict = params_grid(grid_type='energy', readgrid= 0,
-                                grid_cen=(47.266, 83.4265, 54.174),
-                                grid_dim=(46, 30, 37),
-                                grid_int=(0.5, 0.5, 0.5))
+        grid_dict = params_grid(grid_type='energy', readgrid= 2)
         probe_grid_dict = probe_params_grid(probe_id= 0, readgrid= 2, pbfocus= 1, perfill= 0.6,
                                             grid_int=(1.5, 1.5, 1.5))
         params_dict = {
@@ -296,10 +412,7 @@ def params_preset(execution_type: str) -> Dict[str, str]:
         }
 
     elif execution_type == 'docking':
-        grid_dict = params_grid(grid_type='docking', readgrid= 0,
-                                grid_cen=(47.266, 83.4265, 54.174),
-                                grid_dim=(46, 30, 37),
-                                grid_int=(0.5, 0.5, 0.5))
+        grid_dict = params_grid(grid_type='docking', readgrid= 2)
 
         params_dict = {
             'title': 'Docking Mehler Solmajer dielectric',
@@ -326,6 +439,18 @@ def params_preset(execution_type: str) -> Dict[str, str]:
             'tiprot': 5, 'inifoc': 5, 'ninter': 20,
             'clhost': 1, 'minout': 50, 'splitpdb': 0, 'cutelec': 10.0
         }
+    elif execution_type == 'check_only':
+        params_dict = {
+            'title': 'Check_only dry run of CMIP',
+            'CHECKONLY': 1,
+            'readgrid': 2,
+            'calcgrid': 1,
+            'tipcalc': 0,
+            'irest': 0,
+            'ebyatom': 1,
+            'coorfmt': 2,
+            'fvdw': 0.8
+        }
 
     return {**params_dict, **grid_dict, **probe_grid_dict}
 
@@ -399,4 +524,71 @@ def create_params_file(output_params_path: str, input_params_path: str = None,
         for k, v in params_properties_dict.items():
             params_dict[k] = v
 
-    return write_params_file(output_params_path, params_dict)
\ No newline at end of file
+    return write_params_file(output_params_path, params_dict)
+
+def mark_residues(residue_list: Sequence[str], input_cmip_pdb_path: str, output_cmip_pdb_path: str, out_log: logging.Logger = None, global_log: logging.Logger = None) -> None:
+    """Marks using an "X" before the atom type all the residues in *residue_list* and writes the result in *output_cmip_pdb_path*.
+
+        Args:
+            residue_list (list): Residue list in the format "Chain:Resnum" (no spaces between the elements) separated by commas. If empty or none all residues will be marked.
+            local_log (:obj:`logging.Logger`): local log object.
+            global_log (:obj:`logging.Logger`): global log object.
+        """
+    if not residue_list:
+        fu.log(f"Empty residue_list all residues will be marked", out_log, global_log)
+    else:
+        fu.log(f"Residue list: {residue_list}", out_log, global_log)
+
+    with open(input_cmip_pdb_path) as pdb_file_in, open(output_cmip_pdb_path, 'w') as pdb_file_out:
+        residue_set_used = set()
+
+        res_counter = 0
+        for line in pdb_file_in:
+            if _is_atom(line):
+                residue_code = _get_residue_code(line)
+                used_residue = _get_residue_code_in_list(residue_code, residue_list)
+                if not residue_list or used_residue:
+                    res_counter += 1
+                    residue_set_used.add(used_residue)
+                    line = _mark_pdb_atom(line)
+            pdb_file_out.write(line)
+        fu.log(f"{res_counter} residues have been marked", out_log, global_log)
+
+        if residue_list:
+            unused_residues = set(residue_list) - residue_set_used
+            if unused_residues:
+                fu.log(f"The following residues where present in the residue_list and have not been marked: {unused_residues}", out_log, global_log)
+
+def _mark_pdb_atom(line: str) -> str:
+    line = list(line)
+    line.insert(64, 'X')
+    return ''.join(line)
+
+def _get_residue_code_in_list(input_residue, residue_list):
+    if not residue_list:
+        return None
+    for res_code in residue_list:
+        if input_residue == res_code:
+            return res_code
+        chain_rescode, resnum_rescode = res_code.split(":")
+        chain_input, resnum_input = input_residue.split(":")
+        if not chain_rescode:
+            if resnum_rescode == resnum_input:
+                return res_code
+        if not resnum_rescode:
+            if chain_rescode == chain_input:
+                return res_code
+    return None
+
+
+def _get_residue_code(line: str) -> str:
+    return _get_chain(line)+":"+_get_resnum(line)
+
+def _get_chain(line: str) -> str:
+    return line[21].upper()
+
+def _get_resnum(line: str) -> str:
+    return line[22:27].strip()
+
+def _is_atom(line: str) -> str:
+    return line[0:6].strip().upper() in ["ATOM", "HETATM"]
